"""
Copyright (c) 2015 Tim Waugh <tim@cyberelk.net>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""

import argparse
import logging
import os
import sys
from systemd import journal

from journal_brief import (SelectiveReader,
                           LatestJournalEntries,
                           EntryFormatter,
                           JournalFilter)
from journal_brief.config import Config
from journal_brief.constants import PACKAGE, CONFIG_DIR, PRIORITY_MAP


class InstanceConfig(object):
    def __init__(self, config, args):
        self.config = config
        self.args = args

    def get(self, key, default_value=None):
        config_key = key.replace('_', '-')
        try:
            value = getattr(self.args, key)
        except AttributeError:
            value = None
        finally:
            if value is None:
                value = self.config.get(config_key, default_value)

        return value


class CLI(object):
    def __init__(self, args=None):
        self.args = self.get_args(args or sys.argv[1:])
        config = Config(config_file=self.args.conf)
        self.config = InstanceConfig(config, self.args)

    def get_args(self, args):
        description = 'Show new journal entries since last run'
        parser = argparse.ArgumentParser(description=description)
        parser.add_argument('-b', action='store_true', default=False,
                            help='process all entries from the current boot')
        parser.add_argument('-p', '--priority', metavar='PRI',
                            help='show entries at priority PRI and lower',
                            choices=['emerg', 'alert', 'crit', 'err',
                                     'warning', 'notice', 'info', 'debug'])
        parser.add_argument('--conf', metavar='FILE',
                            help='use FILE as config file')
        parser.add_argument('--dry-run', action='store_true', default=False,
                            help='do not update cursor bookmark file')

        cmds = parser.add_subparsers(dest='cmd')
        cmds.add_parser('reset', help='reset cursor bookmark and exit')
        cmds.add_parser('stats', help='show statistics')
        return parser.parse_args(args)

    def show_stats(self, entries, exclusions):
        logging.debug('exclusions: %r', exclusions)
        jfilter = JournalFilter(entries, exclusions=exclusions)
        list(jfilter)
        stats = jfilter.get_statistics()
        logging.debug("stats: %r", stats)
        strf = "{FREQ:>10}  {EXCLUSION}"
        print(strf.format(FREQ='FREQUENCY', EXCLUSION='EXCLUSION'))
        for stat in stats:
            print(strf.format(FREQ=stat.hits, EXCLUSION=repr(stat.exclusion)))

    def run(self):
        cursor_file = self.config.get('cursor_file')
        if not cursor_file.startswith('/'):
            cursor_file = os.path.join(CONFIG_DIR, cursor_file)

        if self.args.cmd == 'reset':
            logging.debug('reset: removing %r', cursor_file)
            try:
                os.unlink(cursor_file)
            except IOError:
                pass

            return

        log_level = None
        priority = self.config.get('priority')
        if priority:
            log_level = int(PRIORITY_MAP[priority])

        formatter = EntryFormatter()
        reader = SelectiveReader(this_boot=self.args.b,
                                 log_level=log_level,
                                 inclusions=self.config.get('inclusions'))
        with LatestJournalEntries(cursor_file=cursor_file,
                                  reader=reader,
                                  dry_run=self.args.dry_run,
                                  seek_cursor=not self.args.b) as entries:
            exclusions = self.config.get('exclusions', [])
            if self.args.cmd == 'stats':
                self.show_stats(entries, exclusions)
            else:
                for entry in JournalFilter(entries, exclusions=exclusions):
                    print(formatter.format(entry))


def run():
    try:
        CLI().run()
    except KeyboardInterrupt:
        pass
    except IOError as ex:
        sys.stderr.write("{0}: {1}\n".format(PACKAGE, os.strerror(ex.errno)))
        sys.exit(1)


if __name__ == '__main__':
    run()



"""
Copyright (c) 2015 Tim Waugh <tim@cyberelk.net>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""

from collections.abc import Iterator
from collections import namedtuple
from journal_brief.constants import PRIORITY_MAP
from logging import getLogger
import os
import re


log = getLogger(__name__)
ExclusionStatistics = namedtuple('ExclusionStatistics', ['hits', 'exclusion'])


# dict, field -> list of values
class Exclusion(dict):
    """
    str (field) -> list (str values)
    """

    def __init__(self, mapping):
        assert isinstance(mapping, dict)

        # Make sure everything is interpreted as a string
        str_mapping = {}
        for field, matches in mapping.items():
            if field == 'PRIORITY':
                str_mapping[field] = [PRIORITY_MAP[match] for match in matches]
            else:
                str_mapping[field] = [str(match) for match in matches]

        super(Exclusion, self).__init__(str_mapping)
        self.hits = 0
        self.regexp = {}  # field -> index -> compiled regexp

    def value_matches(self, field, index, match, value):
        try:
            log.debug('using cached regexp for %s[%d]:%s',
                      field, index, match)
            regexp = self.regexp[field][index]
        except KeyError:
            if match.startswith('/') and match.endswith('/'):
                pattern = match[1:-1]
                log.debug('compiling pattern %r', pattern)
                regexp = re.compile(pattern)
            else:
                regexp = None
                log.debug('%r is not a regex', match)

            self.regexp.setdefault(field, {})
            self.regexp[field][index] = regexp

        if regexp is not None:
            return regexp.match(value)

        return match == value

    def matches(self, entry):
        for field, matches in self.items():
            is_match = False
            for index, match in enumerate(matches):
                if self.value_matches(field, index, match, entry.get(field)):
                    is_match = True
                    break

            if not is_match:
                return False

        self.hits += 1
        return True


class JournalFilter(Iterator):
    """
    Exclude certain journal entries

    Provide a list of exclusions. Each exclusion is a dict whose keys
    are fields which must all match an entry to be excluded.

    The dict value for each field is a list of possible match values,
    any of which may match.

    Regular expressions are indicated with '/' at the beginning and
    end of the match string. Regular expressions are matched at the
    start of the journal field value (i.e. it's a match not a search).
    """

    def __init__(self, iterator, exclusions=None):
        """
        Constructor

        :param iterator: iterator, providing journal entries
        :param exclusions: list, dicts of str(field) -> [str(match), ...]
        """
        super(JournalFilter, self).__init__()
        self.iterator = iterator
        if exclusions:
            self.exclusions = [Exclusion(excl) for excl in exclusions]
        else:
            self.exclusions = []

    def __next__(self):
        for entry in self.iterator:
            if not any(exclusion.matches(entry)
                       for exclusion in self.exclusions):
                return entry

        raise StopIteration

    def get_statistics(self):
        """
        Get filter statistics

        :return: list, ExclusionStatistics instances in reverse order
        """

        stats = [ExclusionStatistics(excl.hits, excl)
                 for excl in self.exclusions]
        stats.sort(reverse=True, key=lambda stat: stat.hits)
        return stats
